@@ -128,6 +128,7 @@ void ScreenCastPortalOutput::setOutputType(const QString &type)
 
 ScreenCastPortal::ScreenCastPortal(QObject *parent)
     : QDBusAbstractAdaptor(parent)
+    , m_eglInitialized(false)
     , m_registryInitialized(false)
     , m_streamingEnabled(false)
     , m_connection(nullptr)
@@ -136,7 +137,6 @@ ScreenCastPortal::ScreenCastPortal(QObject *parent)
     , m_remoteAccessManager(nullptr)
 {
     initDrm();
-    initEGL();
     initWayland();
 
     qDBusRegisterMetaType<ScreenCastPortal::Stream>();
@@ -195,11 +195,20 @@ void ScreenCastPortal::createPipeWireStream(const QSize &resolution)
 void ScreenCastPortal::initDrm()
 {
     m_drmFd = open("/dev/dri/renderD128", O_RDWR);
+
+    if (m_drmFd == -1) {
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Cannot open render node: " << strerror(errno);
+        return;
+    }
+
     m_gbmDevice = gbm_create_device(m_drmFd);
 
     if (!m_gbmDevice) {
-        qFatal("Cannot create GBM device: %s", strerror(errno));
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Cannot create GBM device: " << strerror(errno);
+        return;
     }
+
+    initEGL();
 }
 
 void ScreenCastPortal::initEGL()
@@ -210,7 +219,8 @@ void ScreenCastPortal::initEGL()
     if (clientExtensionsString.isEmpty()) {
         // If eglQueryString() returned NULL, the implementation doesn't support
         // EGL_EXT_client_extensions. Expect an EGL_BAD_DISPLAY error.
-        qFatal("No client extensions defined! %s", formatGLError(eglGetError()));
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "No client extensions defined! " << formatGLError(eglGetError());
+        return;
     }
 
     m_egl.extensions = clientExtensionsString.split(' ');
@@ -219,32 +229,39 @@ void ScreenCastPortal::initEGL()
     // if the implementation supports it.
     if (!m_egl.extensions.contains(QByteArrayLiteral("EGL_EXT_platform_base")) ||
             !m_egl.extensions.contains(QByteArrayLiteral("EGL_MESA_platform_gbm"))) {
-        qFatal("One of required EGL extensions is missing");
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "One of required EGL extensions is missing";
+        return;
     }
 
     m_egl.display = eglGetPlatformDisplayEXT(EGL_PLATFORM_GBM_MESA, m_gbmDevice, nullptr);
 
     if (m_egl.display == EGL_NO_DISPLAY) {
-        qFatal("Error during obtaining EGL display: %s", formatGLError(eglGetError()));
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Error during obtaining EGL display: " << formatGLError(eglGetError());
+        return;
     }
 
     EGLint major, minor;
     if (eglInitialize(m_egl.display, &major, &minor) == EGL_FALSE) {
-        qFatal("Error during eglInitialize: %s", formatGLError(eglGetError()));
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Error during eglInitialize: " << formatGLError(eglGetError());
+        return;
     }
 
     if (eglBindAPI(EGL_OPENGL_API) == EGL_FALSE) {
-        qFatal("bind OpenGL API failed");
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "bind OpenGL API failed";
+        return;
     }
 
     m_egl.context = eglCreateContext(m_egl.display, nullptr, EGL_NO_CONTEXT, nullptr);
 
     if (m_egl.context == EGL_NO_CONTEXT) {
-        qFatal("Couldn't create EGL context: %s", formatGLError(eglGetError()));
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Couldn't create EGL context: " << formatGLError(eglGetError());
+        return;
     }
 
     qCDebug(XdgDesktopPortalKdeScreenCast) << "Egl initialization succeeded";
     qCDebug(XdgDesktopPortalKdeScreenCast) << QString("EGL version: %1.%2").arg(major).arg(minor);
+
+    m_eglInitialized = true;
 }
 
 void ScreenCastPortal::initWayland()
@@ -379,12 +396,16 @@ uint ScreenCastPortal::Start(const QDBusObjectPath &handle,
         return 2;
     }
 
-    // TODO check whether we got some outputs?
     if (m_outputMap.isEmpty()) {
         qCWarning(XdgDesktopPortalKdeScreenCast) << "Failed to show dialog as there is no screen to select";
         return 2;
     }
 
+    if (!m_eglInitialized) {
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "EGL is not initialized, we are not able to process screen content";
+        return 2;
+    }
+
     QScopedPointer<ScreenChooserDialog, QScopedPointerDeleteLater> screenDialog(new ScreenChooserDialog(m_outputMap, session->multipleSources()));
 
     if (screenDialog->exec()) {
@@ -427,7 +448,7 @@ uint ScreenCastPortal::Start(const QDBusObjectPath &handle,
         return 0;
     }
 
-    return 0;
+    return 1;
 }
 
 void ScreenCastPortal::addOutput(quint32 name, quint32 version)
@@ -481,15 +502,25 @@ void ScreenCastPortal::processBuffer(const KWayland::Client::RemoteBuffer* rbuf)
         return;
     }
 
+    if (m_lastFrameTime.isValid() &&
+        m_lastFrameTime.msecsTo(QDateTime::currentDateTime()) < (1000 / m_stream->framerate())) {
+        close(gbmHandle);
+        return;
+    }
+
     if (!gbm_device_is_format_supported(m_gbmDevice, format, GBM_BO_USE_SCANOUT)) {
-        qCritical() << "GBM format is not supported by device!";
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Failed to process buffer: GBM format is not supported by device!";
+        close(gbmHandle);
+        return;
     }
 
     // import GBM buffer that was passed from KWin
     gbm_import_fd_data importInfo = {gbmHandle, width, height, stride, format};
     gbm_bo *imported = gbm_bo_import(m_gbmDevice, GBM_BO_IMPORT_FD, &importInfo, GBM_BO_USE_SCANOUT);
     if (!imported) {
-        qCritical() << "Cannot import passed GBM fd:" << strerror(errno);
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Failed to process buffer: Cannot import passed GBM fd - " << strerror(errno);
+        close(gbmHandle);
+        return;
     }
 
     // bind context to render thread
@@ -497,10 +528,16 @@ void ScreenCastPortal::processBuffer(const KWayland::Client::RemoteBuffer* rbuf)
 
     // create EGL image from imported BO
     EGLImageKHR image = eglCreateImageKHR(m_egl.display, nullptr, EGL_NATIVE_PIXMAP_KHR, imported, nullptr);
+
+    // We can already close gbm handle
+    gbm_bo_destroy(imported);
+    close(gbmHandle);
+
     if (image == EGL_NO_IMAGE_KHR) {
-        qCritical() << "Error creating EGLImageKHR" << formatGLError(glGetError());
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Failed to process buffer: Error creating EGLImageKHR - " << formatGLError(glGetError());
         return;
     }
+
     // create GL 2D texture for framebuffer
     GLuint texture;
     glGenTextures(1, &texture);
@@ -518,7 +555,7 @@ void ScreenCastPortal::processBuffer(const KWayland::Client::RemoteBuffer* rbuf)
     glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture, 0);
     const GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
     if (status != GL_FRAMEBUFFER_COMPLETE) {
-        qCritical() << "glCheckFramebufferStatus failed:" << formatGLError(glGetError());
+        qCWarning(XdgDesktopPortalKdeScreenCast) << "Failed to process buffer: glCheckFramebufferStatus failed - " << formatGLError(glGetError());
         glDeleteTextures(1, &texture);
         glDeleteFramebuffers(1, &framebuffer);
         eglDestroyImageKHR(m_egl.display, image);
@@ -529,16 +566,15 @@ void ScreenCastPortal::processBuffer(const KWayland::Client::RemoteBuffer* rbuf)
     glViewport(0, 0, width, height);
     glGetTexImage(GL_TEXTURE_2D, 0, GL_RGBA, GL_UNSIGNED_BYTE, capture->bits());
 
-    m_stream->recordFrame(capture->bits());
+    if (m_stream->recordFrame(capture->bits())) {
+        m_lastFrameTime = QDateTime::currentDateTime();
+    }
 
-    gbm_bo_destroy(imported);
     glDeleteTextures(1, &texture);
     glDeleteFramebuffers(1, &framebuffer);
     eglDestroyImageKHR(m_egl.display, image);
 
     delete capture;
-
-    close(gbmHandle);
 }
 
 void ScreenCastPortal::setupRegistry()
